from .waveform import (plot_waveform2D,plot_waveform_wiggle,plot_waveform_trace)

from .velocity_model import (plot_vp_rho,plot_vp_vs_rho,plot_eps_delta_gamma,plot_lam_mu,plot_model)

from .boundary_condition import plot_bcx_bcz,plot_damp

from .survey import plot_survey,plot_wavelet

from .inverted_loss_model import *